package pku.edu.hocra.graph;

import pku.edu.hocra.graph.ds.IncrementalIndexing;
import pku.edu.hocra.graph.ds.IntPair;
import pku.edu.hocra.graph.ds.MutualMap;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import soot.Kind;
import soot.SootMethod;
import soot.Unit;
import soot.jimple.toolkits.callgraph.CallGraph;
import soot.jimple.toolkits.callgraph.Edge;

public class ExplodedGraphBuilder {
	
	private CallGraph cg;
	private Set<SootMethod> methods;
	private IncrementalIndexing<String> mNameId;
	private MutualMap<String, ProceduralGraph> mName2pg;
	private Map<String, Integer> baseIndex;
	
	public ExplodedGraphBuilder(CallGraph cg, Set<SootMethod> methods) {
		this.cg = cg;
		this.methods = methods;
		this.mNameId = new IncrementalIndexing<String>();
		this.mName2pg = new MutualMap<String, ProceduralGraph>();
		this.baseIndex = new HashMap<String, Integer>();
	}
	
	public String getSootMethodSignature(SootMethod sm) {
		return sm.getSignature() + sm.getActiveBody().toString().hashCode();
	}
	
	public void printCallGraph(String cgFileName) throws FileNotFoundException {
		FileOutputStream fos = new FileOutputStream(cgFileName);
		PrintStream ps = new PrintStream(fos);
		for (Edge e : cg) {
			if (e.src().hasActiveBody() && e.tgt().hasActiveBody() && e.srcUnit() != null
					&& methods.contains(e.src()) && methods.contains(e.tgt())
					&& !e.src().getSignature().contains("<clinit>") && mName2pg.getValue(e.src().toString()).getNodeNumber() > 0
					&& !e.tgt().getSignature().contains("<clinit>") && mName2pg.getValue(e.tgt().toString()).getNodeNumber() > 0
					&& (e.kind() == Kind.INTERFACE || e.kind() == Kind.VIRTUAL || e.kind() == Kind.SPECIAL || e.kind() == Kind.STATIC)) {
				ps.println(mNameId.getIndex(e.src().getSignature()) + " " 
						+ getUnitIndex(e.srcUnit(), e.src()) + " " 
						+ mNameId.getIndex(e.tgt().getSignature()));
			}
		}
		ps.close();
	}

	public void build(String methodFileName, String graphFileName, 
			String cgFileName, String maxnumFileName) throws FileNotFoundException {
		int m = 0;
		int cb = 0;
		int n = 0;
		
		FileOutputStream fos = new FileOutputStream(graphFileName);
		PrintStream ps = new PrintStream(fos);
		FileOutputStream fosCg = new FileOutputStream(cgFileName);
		PrintStream psCg = new PrintStream(fosCg);
		
		for (SootMethod sm : methods) {
			if (!sm.hasActiveBody() || sm.getSignature().contains("<clinit>")) continue;
			mNameId.append(sm.getSignature());
			ProceduralGraph pg = new ProceduralGraph(sm);
			mName2pg.put(sm.getSignature(), pg);
			baseIndex.put(sm.getSignature(), n);
			for (IntPair ip : pg.getEdges()) {
				ps.println((ip.u+n) + " " + (ip.v+n));
			}
			for (IntPair cp : pg.callbackNodePairSet) {
				ps.println((cp.u+n) + " " + (cp.v+n) + " C " + (cb++));
			}
			n += pg.getNodeNumber(); // cannot be deleted
			m += pg.getEdges().size();
		}
		
		if (maxnumFileName != null) {
			PrintStream psLibSep = new PrintStream(new FileOutputStream(maxnumFileName)); 
			psLibSep.println(n+2);
			psLibSep.close();
		}
			
		int m1 = 0, m2 = 0, vtc = 0, pi = 0;
		for (Edge e : cg) {
			if (e.src().hasActiveBody() && e.tgt().hasActiveBody() && e.srcUnit() != null
					&& methods.contains(e.src()) && methods.contains(e.tgt())
					&& !e.src().getSignature().contains("<clinit>")
					&& !e.tgt().getSignature().contains("<clinit>")
					&& (e.kind() == Kind.INTERFACE || e.kind() == Kind.VIRTUAL || e.kind() == Kind.SPECIAL || e.kind() == Kind.STATIC)) {
				int ci = getUnitIndex(e.srcUnit(), e.src());
				pi ++;
				ProceduralGraph srcPg = mName2pg.getValue(e.src().toString());
				ProceduralGraph tgtPg = mName2pg.getValue(e.tgt().toString());
				Map<Integer, HashMap<Integer, Integer>> callp = srcPg.callStmtPrmt2Local;
				int srcBase = baseIndex.get(e.src().toString());
				int tgtBase = baseIndex.get(e.tgt().toString());
				int newvtc = 0;
				if (callp.containsKey(ci)) {
					HashMap<Integer, Integer> h = callp.get(ci);
					for (Entry<Integer, Integer> et : h.entrySet()) {
						if (tgtPg.entryPrmt2Local.containsKey(et.getKey())) {
							m1 ++;
							if (e.kind() == Kind.INTERFACE || e.kind() == Kind.VIRTUAL) {
								newvtc = 1;
								ps.println((et.getValue()+srcBase) + " " + (tgtPg.entryPrmt2Local.get(et.getKey())+tgtBase) + " LV " + pi);
							}
							else
								ps.println((et.getValue()+srcBase) + " " + (tgtPg.entryPrmt2Local.get(et.getKey())+tgtBase) + " L " + pi);
							// et.getValue() -> tgtPg.entryPrmt2Local.get(et.getKey())
						}
					}
				}
				Map<Integer, Integer> returnp = srcPg.retStmt2Local;
				if (returnp.containsKey(ci)) {
					Integer r = returnp.get(ci);
					for (Integer x : tgtPg.exitLocals) {
						m2 ++;
						if (e.kind() == Kind.INTERFACE || e.kind() == Kind.VIRTUAL) {
							newvtc = 1;
							ps.println((x+tgtBase) + " " + (r+srcBase) + " RV " + pi);
						}
						else
							ps.println((x+tgtBase) + " " + (r+srcBase) + " R " + pi);
					}
				}
				vtc += newvtc;
				psCg.print(mNameId.getIndex(e.src().getSignature()) + " " 
						+ getUnitIndex(e.srcUnit(), e.src()) + " " 
						+ mNameId.getIndex(e.tgt().getSignature()));
				if (e.kind() == Kind.INTERFACE || e.kind() == Kind.VIRTUAL)
					psCg.println(" V " + pi);
				else
					psCg.println(" N " + pi);
			}
		}
		System.out.println(m + "; " + m1 + "; " + m2 + "; " + cb + "; " + vtc);
		ps.close();
		psCg.close();
		
		FileOutputStream fosMethod = new FileOutputStream(methodFileName);
		PrintStream psMethod = new PrintStream(fosMethod);
		for (SootMethod sm : methods) {
			if (sm.hasActiveBody() && !sm.toString().contains("<clinit>") && mName2pg.getValue(sm.toString()).getNodeNumber() > 0) {
				psMethod.println(getSootMethodSignature(sm) 
						+ ";" + baseIndex.get(sm.toString()) 
						+ ";" + mName2pg.getValue(sm.toString()).getNodeNumber() 
						+ ";" + mNameId.getIndex(sm.getSignature()));
			}
		}
		psMethod.close();
		
		//printCallGraph(cgFileName);
		
	}
	
	private int getUnitIndex(Unit u0, SootMethod m) {
		int i = 0;
		for (Unit u : m.getActiveBody().getUnits()) {
			if (u0.equals(u))
				return i;
			i++;
		}
		return -1;
	}
	
}
