//
// Created by Wayne on 16/1/23.
//

#include "tal_library_graph.h"

using std::ifstream;
using std::stringstream;
using std::getline;
using std::make_pair;

vector<const tal_fst_edge*> tal_library_graph::FST_EMPTY;
vector<const tal_snd_edge*> tal_library_graph::SND_EMPTY;

tal_library_graph::tal_library_graph(int num_nodes) {
    this->num_nodes = num_nodes;
    this->num_edges = 0;
    for (int s = 0; s < 7; s++) {
        this->fst_forward[s] = new vector<const tal_fst_edge*>[num_nodes];
        this->fst_backward[s] = new vector<const tal_fst_edge*>[num_nodes];
        this->fst_record[s] = new unordered_set<int>[num_nodes];
    }
    for (int s = 0; s < 3; s++) {
        this->snd_forward[s] = new vector<const tal_snd_edge*>[num_nodes];
        this->snd_backward[s] = new vector<const tal_snd_edge*>[num_nodes];
        this->snd_forward_mid[s] = new vector<const tal_snd_edge*>[num_nodes];
        this->snd_backward_mid[s] = new vector<const tal_snd_edge*>[num_nodes];
        this->snd_of[s] = new unordered_map<int, vector<const tal_snd_edge*>>[num_nodes];
        this->snd_of_mid[s] = new unordered_map<int, vector<const tal_snd_edge*>>[num_nodes];
        this->snd_record[s] = new unordered_map<int, unordered_map<int, unordered_set<int>>>[num_nodes];
    }
}

int tal_library_graph::get_num_nodes() const {
    return this->num_nodes;
}

int tal_library_graph::get_num_edges() const {
    return this->num_edges;
}

void tal_library_graph::add_paren(const tal_fst_edge *e) {
    if (is_left_paren(e->get_symbol())) {
        this->left_parens[e->get_symbol()].push_back(e);
        this->call_nodes.insert(e->get_source());
        this->entry_nodes.insert(e->get_target());
    } else if (is_right_paren(e->get_symbol())) {
        this->right_parens[e->get_symbol()].push_back(e);
        this->exit_nodes.insert(e->get_source());
        this->return_nodes.insert(e->get_target());
    } else {
        throw "bad implementation";
    }
    this->parens.push_back(e);
}

void tal_library_graph::add_fst_edge(const tal_fst_edge *fe) {
    this->num_edges += 1;
    this->fst_forward[fe->get_symbol()][fe->get_source()].push_back(fe);
    this->fst_backward[fe->get_symbol()][fe->get_target()].push_back(fe);
}

void tal_library_graph::add_snd_edge(const tal_snd_edge *se) {
    this->num_edges += 1;
    this->snd_forward[se->get_symbol()][se->get_source()].push_back(se);
    this->snd_backward[se->get_symbol()][se->get_target()].push_back(se);
    this->snd_forward_mid[se->get_symbol()][se->get_mid_source()].push_back(se);
    this->snd_backward_mid[se->get_symbol()][se->get_mid_target()].push_back(se);
    this->snd_of[se->get_symbol()][se->get_source()][se->get_target()].push_back(se);
    this->snd_of_mid[se->get_symbol()][se->get_mid_source()][se->get_mid_target()].push_back(se);
}

const vector<const tal_fst_edge*>& tal_library_graph::fst_edges_from(int source, int symbol) const {
    return this->fst_forward[symbol][source];
}

const vector<const tal_fst_edge*>& tal_library_graph::fst_edges_to(int target, int symbol) const {
    return this->fst_backward[symbol][target];
}

const vector<const tal_snd_edge*>& tal_library_graph::snd_edges_from(int source, int symbol) const {
    return this->snd_forward[symbol][source];
}

const vector<const tal_snd_edge*>& tal_library_graph::snd_edges_to(int target, int symbol) const {
    return this->snd_backward[symbol][target];
}

const vector<const tal_snd_edge*>& tal_library_graph::snd_edges_from_mid(int source, int symbol) const {
    return this->snd_forward_mid[symbol][source];
}

const vector<const tal_snd_edge*>& tal_library_graph::snd_edges_to_mid(int target, int symbol) const {
    return this->snd_backward_mid[symbol][target];
}

const vector<const tal_snd_edge*>& tal_library_graph::snd_edges_of(int source, int target, int symbol) const {
    if (this->snd_of[symbol][source].count(target) != 0) {
        return this->snd_of[symbol][source].at(target);
    } else {
        return SND_EMPTY;
    }
}

const vector<const tal_snd_edge*>& tal_library_graph::snd_edges_of_mid(int source, int target, int symbol) const {
    if (this->snd_of_mid[symbol][source].count(target) != 0) {
        return this->snd_of_mid[symbol][source].at(target);
    } else {
        return SND_EMPTY;
    }
}

const vector<const tal_fst_edge*>& tal_library_graph::all_left_parens_of(int symbol) const {
    if (this->left_parens.count(symbol) != 0) {
        return this->left_parens.at(symbol);
    } else {
        return FST_EMPTY;
    }
}

const vector<const tal_fst_edge*>& tal_library_graph::all_right_parens_of(int symbol) const {
    if (this->right_parens.count(symbol) != 0) {
        return this->right_parens.at(symbol);
    } else {
        return FST_EMPTY;
    }
}

void tal_library_graph::add_fst_record(const tal_fst_edge *fe) {
    this->fst_record[fe->get_symbol()][fe->get_source()].insert(fe->get_target());
}

void tal_library_graph::add_snd_record(const tal_snd_edge *se) {
    this->snd_record[se->get_symbol()][se->get_source()][se->get_target()][se->get_mid_source()].insert(se->get_mid_target());
}

bool tal_library_graph::is_fst_redundant(int source, int target, int symbol) const {
    return (this->fst_record[symbol][source].count(target) != 0);
}

bool tal_library_graph::is_snd_redundant(int source, int target, int mid_source, int mid_target, int symbol) const {
    if (this->snd_record[symbol][source].count(target) != 0) {
        auto &t1 = this->snd_record[symbol][source].at(target);
        if (t1.count(mid_source) != 0) {
            auto &t2 = t1.at(mid_source);
            return (t2.count(mid_target) != 0);
        } else {
            return false;
        }
    } else {
        return false;
    }
}

const vector<const tal_fst_edge*>& tal_library_graph::all_parens() const {
    return this->parens;
}

tal_library_graph* tal_library_graph::load2(const string &file_name) {
    ifstream fin(file_name);
    int num_nodes;
    fin >> num_nodes;
    tal_library_graph *lib_gr = new tal_library_graph(num_nodes);
    string line;
    getline(fin, line);
    while (getline(fin, line)) {
        stringstream sin(line);
        int source, target;
        sin >> source >> target;
        string type;
        if (sin >> type) {
            int id;
            sin >> id;
            if (type[0] == 'C') {
            } else if (type[0] == 'L') {
                id += 10;
                lib_gr->add_paren(new tal_fst_edge(id, source, target));
            } else if (type[0] == 'R') {
                id += 10;
                lib_gr->add_paren(new tal_fst_edge(-id, source, target));
            } else {
                throw "bad implementation";
            }
        } else {
            lib_gr->add_fst_edge(new tal_fst_edge(GRAM_S, source, target));
        }
    }
    return lib_gr;
}

const unordered_map<int, vector<const tal_fst_edge*>>& tal_library_graph::get_left_parens() const {
    return this->left_parens;
}

const unordered_map<int, vector<const tal_fst_edge*>>& tal_library_graph::get_right_parens() const {
    return this->right_parens;
}

bool tal_library_graph::is_call_node(int n) const {
    return (this->call_nodes.count(n) != 0);
}

bool tal_library_graph::is_return_node(int n) const {
    return (this->return_nodes.count(n) != 0);
}

bool tal_library_graph::is_entry_node(int n) const {
    return (this->entry_nodes.count(n) != 0);
}

bool tal_library_graph::is_exit_node(int n) const {
    return (this->exit_nodes.count(n) != 0);
}

void tal_library_graph::add_chain_node(int n) {
    this->chain_nodes.insert(n);
}

void tal_library_graph::load_lib_info(const string &file_name) {
    ifstream fin(file_name);
    for (int start, len; fin >> start >> len; this->lib_info.push_back(make_pair(start, len)));
    fin.close();
}

const vector<pair<int, int>>& tal_library_graph::get_lib_info() const {
    return this->lib_info;
}

bool tal_library_graph::is_chain_node(int n) const {
    return (this->chain_nodes.count(n) != 0);
}

const unordered_set<int>& tal_library_graph::get_chain_nodes() const {
    return this->chain_nodes;
}